from dateutil.relativedelta import relativedelta


class WeekExpirationMixin:
    expiration = relativedelta(weeks=1)


class MonthExpirationMixin:
    expiration = relativedelta(months=1)


class QuarterExpirationMixin:
    expiration = relativedelta(months=3)


class YearExpirationMixin:
    expiration = relativedelta(years=1)



from rest_framework.permissions import BasePermission


class CanCreateTaskPermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.is_superuser or request.user.has_perm('gonk.can_create_task')


class CanCancelTaskPermission(BasePermission):
    def has_object_permission(self, request, view, obj):
        return request.user.is_superuser or request.user.has_perm('gonk.can_cancel_task') and request.user.username == obj.username


class CanRevertTaskPermission(BasePermission):
    def has_object_permission(self, request, view, obj):
        if not obj.get_taskrunner().reversible:
            return False

        return request.user.is_superuser or request.user.has_perm('gonk.can_revert_task') and request.user.username == obj.username



import json

try:
    from django_celery_beat.models import CrontabSchedule, PeriodicTask
except ImportError as e:
    raise ImportError(
        'Required dependency django_celery_beat is not installed. '
        'Perhaps you want to install persistance extra dependencies. '
    ) from e


def add_persistent_celery_beat(name: str, task: str, schedule: CrontabSchedule, *args):
    PeriodicTask.objects.create(
        name=name,
        task='gonk.tasks.run_schedule',
        crontab=schedule,
        args=json.dumps([task, *args]),
    )



from rest_framework.routers import DefaultRouter

from gonk.contrib.rest_framework.viewsets import TaskViewSet

router = DefaultRouter(trailing_slash=True)

router.register(r'', TaskViewSet, basename='')

urlpatterns = router.urls



from django.conf import settings


MERCURE_HUB_URL = getattr(settings, 'MERCURE_HUB_URL', '')
MERCURE_JWT_KEY = getattr(settings, 'MERCURE_JWT_KEY', '')

DEFAULT_NOTIFICATION_EMAIL = getattr(settings, 'DEFAULT_NOTIFICATION_EMAIL', '')



from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet
from rest_framework import mixins

from gonk.contrib.rest_framework.exceptions import APIValidationException
from gonk.exceptions import TaskRunnerValidationException
from gonk.models import Task
from gonk.contrib.rest_framework import permissions
from gonk.contrib.rest_framework import serializers


class TaskViewSet(mixins.CreateModelMixin, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet):
    serializer_class = serializers.TaskSerializer
    queryset = Task.objects.all()
    permission_classes = (IsAuthenticated, )

    def get_queryset(self):
        if self.request.user.is_superuser:
            return super().get_queryset()

        return super().get_queryset().filter(username=self.request.user.username)

    def get_serializer_class(self):
        if self.action == 'create':
            return serializers.CreateTaskSerializer
        if self.action == 'retrieve':
            return serializers.RetrieveTaskSerializer

        return self.serializer_class

    def get_permissions(self):
        if self.action == 'create':
            return [IsAuthenticated(), permissions.CanCreateTaskPermission()]

        return super().get_permissions()

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        task = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        serializer = self.serializer_class(instance=task)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        try:
            validated_data = serializer.validated_data
            user = self.request.user
            params = {'username': user.username, **validated_data}
            return Task.create_task(**params)
        except TaskRunnerValidationException as e:
            raise APIValidationException(str(e))

    @action(methods=['put', 'patch'], detail=True, permission_classes=[
        IsAuthenticated,
        permissions.CanRevertTaskPermission
    ])
    def revert(self, request, *args, **kwargs) -> Response:
        task = self.get_object()
        task.revert()
        serializer = self.get_serializer_class()
        serializer = serializer(instance=task)
        return Response(data=serializer.data, status=status.HTTP_200_OK)

    @action(detail=True, methods=['post'], permission_classes=[
        IsAuthenticated,
        permissions.CanCancelTaskPermission
    ])
    def cancel(self, request, pk, *args, **kwargs) -> Response:
        task = self.get_object()
        task.cancel()
        serializer = self.get_serializer_class()
        serializer = serializer(instance=task)
        return Response(data=serializer.data, status=status.HTTP_200_OK)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _


STATUS_ERROR = 'ERROR'
STATUS_PENDING = 'PENDI'
STATUS_DOING = 'DOING'
STATUS_DONE = 'DONE'
STATUS_CANCELLING = 'CANCG'
STATUS_CANCELED = 'CANCD'
STATUS_TO_REVERT = 'TOREV'
STATUS_REVERTING = 'REVTG'
STATUS_REVERTED = 'REVTD'
STATUS_RETRYING = 'RTRNG'
STATUS_RETRY_ERROR = 'RTERR'


class TaskStatusChoices(models.TextChoices):
    ERROR = STATUS_ERROR, _('Error')
    PENDING = STATUS_PENDING, _('To DO')
    DOING = STATUS_DOING, _('Doing')
    DONE = STATUS_DONE, _('Done')
    CANCELLING = STATUS_CANCELLING, _('Cancelling')
    CANCELED = STATUS_CANCELED, _('Canceled')
    TO_REVERT = STATUS_TO_REVERT, _('To Revert')
    REVERTING = STATUS_REVERTING, _('Reverting')
    REVERTED = STATUS_REVERTED, _('Reverted')
    RETRYING = STATUS_RETRYING, _('Retrying')
    RETRY_ERROR = STATUS_RETRY_ERROR, _('Retry error')


OK = u'OK'
KO = u'KO'
WARNING = u'WA'
TASK_RESULT = (
    (OK, u'OK'),
    (KO, u'ERROR'),
    (WARNING, u'WARNING'),
)



from typing import Optional

from dateutil.relativedelta import relativedelta


class BaseTaskRunner:
    expiration: Optional[relativedelta] = None
    reversible: bool = True

    def __init__(self, task):
        self.task = task

    def validate(self):
        """
        :raises gonk.exceptions.TaskRunnerValidationException:
        """
        raise NotImplementedError()

    def run(self):
        raise NotImplementedError()

    def revert(self):
        raise NotImplementedError()

    def notify(self, data: str):
        raise NotImplementedError()

    def retry(self):
        raise NotImplementedError()


class TaskRunner(BaseTaskRunner):

    def run(self):
        pass

    def revert(self):
        pass

    def notify(self, data: str):
        pass

    def retry(self):
        return self.run()

    def expire(self):
        pass



from gonk.registry import REGISTRY
from celery.schedules import crontab


def register(name):
    def register_class(cls):
        REGISTRY.register(name, cls)
        return cls

    return register_class


def register_beat(name, cron: crontab):
    def register_class(cls):
        REGISTRY.register_beat(name, cls, cron)
        return cls

    return register_class



from threading import Thread

import jwt
import requests
from gonk.contrib.notifications import settings


def make_request(headers, data):
    return requests.post(
        settings.MERCURE_HUB_URL,
        data=data,
        headers=headers,
    )


def publish_event(topic: str, targets: list, data: str) -> 'requests.Response':
    """
    Publishes an event to the Mercury Hub.
    topic: The topic the event will be sent to. Only subscribers who request this topic will get notified.
    targets: The targets that are eligible to get the event.
    data: The data to publish
    """
    token = get_jwt_token([], targets)
    headers = {
        'Authorization': 'Bearer {}'.format(token),
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    data = {
        'topic': topic,
        'data': data,
        'private': 'on'
    }

    return requests.post(
        settings.MERCURE_HUB_URL,
        data=data,
        headers=headers,
    )


def get_jwt_token(subscribe_targets: list, publish_targets: list) -> str:
    """
    Creates a Mercure JWT token with the subscribe and publish targets.
    The JWT token gets signed with a key shared with the Mercure Hub.
    """
    return jwt.encode(
        {
            'mercure': {
                'subscribe': subscribe_targets,
                'publish': publish_targets
            }
        },
        settings.MERCURE_JWT_KEY,
        algorithm='HS256'
    ).decode('utf-8')


class MercureNotificationMixin:
    def get_username(self):
        return self.task.username if self.task.username else settings.DEFAULT_NOTIFICATION_EMAIL

    def get_topic_name(self):
        username = self.get_username()
        return f'gonk-event-{username}-{self.task.celery_id}'

    def notify(self, data):
        if settings.MERCURE_HUB_URL:
            username = self.get_username()
            topic = self.get_topic_name()

            t = Thread(target=publish_event, args=(topic, [topic, username], data))
            t.start()

        super(MercureNotificationMixin, self).notify(data)



# Generated by Django 3.2 on 2022-07-04 10:24

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('celery_id', models.CharField(max_length=120)),
                ('runner_path', models.CharField(max_length=255)),
                ('input', models.JSONField(default=dict)),
                ('results', models.JSONField(default=dict)),
                ('username', models.CharField(default='', max_length=100, verbose_name='Username')),
                ('started_on', models.DateTimeField(blank=True, null=True)),
                ('finished_on', models.DateTimeField(blank=True, null=True)),
                ('revert_started_on', models.DateTimeField(blank=True, null=True)),
                ('revert_finished_on', models.DateTimeField(blank=True, null=True)),
                ('log', models.TextField(default='')),
                ('status', models.CharField(choices=[('ERROR', 'Error'), ('PENDI', 'To DO'), ('DOING', 'Doing'), ('DONE', 'Done'), ('CANCG', 'Cancelling'), ('CANCD', 'Canceled'), ('TOREV', 'To Revert'), ('REVTG', 'Reverting'), ('REVTD', 'Reverted')], default='PENDI', max_length=5)),
                ('expire_on', models.DateTimeField(blank=True, null=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('queue', models.CharField(default='celery', max_length=32)),
            ],
            options={
                'permissions': (('can_create_task', 'Can create task'), ('can_revert_task', 'Can revert task'), ('can_cancel_task', 'Can cancel task')),
            },
        ),
    ]



from gonk.beat import add_beat_to_celery


class TaskRegistry:
    def __init__(self):
        self.registry = {}

    def register(self, name, taskrunner):
        self.registry[name] = f"{taskrunner.__module__}.{taskrunner.__qualname__}"

    def register_beat(self, name, taskrunner, cron):
        self.registry[name] = f"{taskrunner.__module__}.{taskrunner.__qualname__}"

        add_beat_to_celery(name, 'gonk.tasks.run_schedule', cron, [name])


REGISTRY = TaskRegistry()



import argparse
import datetime
import json

from django.core.management.base import BaseCommand
from django.utils import timezone

from gonk.models import Task


class Command(BaseCommand):
    """
        Usage:

        python manage.py create_task <task_type> --raw-input='{}'
        cat file.json | python manage.py create_task <task_type> --queue="celery" --input -
    """
    def add_arguments(self, parser):
        parser.add_argument(
            'task_type',
            type=str,
            help=f'Task type identifier'
        )
        parser.add_argument(
            '--input',
            type=argparse.FileType('r'),
            required=False,
            help='File input -- can be redirected from standard output'
        )
        parser.add_argument(
            '--raw-input',
            type=str,
            required=False,
            help='Raw string input -- Must be in json format'
        )
        parser.add_argument(
            '--queue',
            type=str,
            required=False,
            default='celery',
            help='Celery queue name in which the task will be run'
        )
        parser.add_argument(
            '--when',
            type=str,
            required=False,
            help='Scheduled task run date -- ISO Format'
        )

    def handle(self, task_type, input, raw_input, queue, when, *args, **options):
        data = None
        eta = timezone.now()

        if input:
            data = json.loads(input.read())

        if raw_input:
            data = json.loads(raw_input)

        if when:
            eta = datetime.datetime.fromisoformat(when)

        Task.create_task(
            task_type=task_type,
            task_input=data,
            queue=queue,
            eta=eta,
        )

        self.stdout.write(self.style.SUCCESS(f'Created task to run on {eta}'))



import importlib
from datetime import datetime, timedelta
from typing import Type, Optional

from django.db import models
from django.utils import timezone
from django.utils.text import gettext_lazy as _

from gonk.registry import REGISTRY
from gonk.settings import TaskStatusChoices
from gonk.taskrunners import TaskRunner


class Task(models.Model):
    celery_id = models.CharField(max_length=120)
    runner_path = models.CharField(max_length=255)
    input = models.JSONField(default=dict)

    results = models.JSONField(default=dict)
    username = models.CharField('Username', max_length=100, default='')
    started_on = models.DateTimeField(null=True, blank=True)
    finished_on = models.DateTimeField(null=True, blank=True)
    revert_started_on = models.DateTimeField(null=True, blank=True)
    revert_finished_on = models.DateTimeField(null=True, blank=True)

    log = models.TextField(default='')
    status = models.CharField(max_length=5,
                              choices=TaskStatusChoices.choices,
                              default=TaskStatusChoices.PENDING)
    expire_on = models.DateTimeField(null=True, blank=True)
    modified = models.DateTimeField(auto_now=True)
    created = models.DateTimeField(auto_now_add=True)

    queue = models.CharField(max_length=32, default='celery')

    retryable = models.BooleanField(default=False)
    retry_time = models.DurationField(default=timedelta(seconds=0), null=True, blank=True)
    retries = models.PositiveIntegerField(default=0)
    max_retries = models.PositiveIntegerField(default=3)

    class Meta:
        permissions = (
            ('can_create_task', _('Can create task')),
            ('can_revert_task', _('Can revert task')),
            ('can_cancel_task', _('Can cancel task')),
        )

    @classmethod
    def create_task(
            cls,
            task_type,
            task_input,
            username='',
            eta: Optional[datetime] = None,
            queue='celery',
            retryable: bool = False,
            retry_seconds: int = 0,
            max_retries: int = 0
    ):
        runner_path = REGISTRY.registry.get(task_type)

        retry_time = timedelta(seconds=retry_seconds)

        task = cls(
            runner_path=runner_path,
            input=task_input,
            queue=queue,
            username=username,
            retryable=retryable,
            retry_time=retry_time,
            max_retries=max_retries,
        )

        expiration = task.get_taskrunner().expiration
        task.expire_on = timezone.now() + expiration if expiration else None
        task.run(eta=eta)
        return task

    @classmethod
    def cleanup(cls):
        for t in Task.objects.filter(expire_on__lte=timezone.now()):
            t.expire()
            t.delete()

    def get_taskrunner_class(self) -> Type[TaskRunner]:
        class_module, class_name = self.runner_path.rsplit('.', 1)
        module = importlib.import_module(class_module)
        return getattr(module, class_name)

    def get_taskrunner(self) -> TaskRunner:
        taskrunner = self.get_taskrunner_class()
        return taskrunner(self)

    def validate(self) -> bool:
        return self.get_taskrunner().validate()

    def run(self, eta: Optional[datetime] = None):
        from gonk.tasks import to_run

        self.validate()

        if not self.id:
            self.save()

        celery_task = to_run.apply_async(queue=self.queue, kwargs={'task_id': self.id}, eta=eta)
        self.celery_id = celery_task.id

        self.save(update_fields=['celery_id'])

    def retry(self):
        if not self.retryable:
            return

        if self.retries > self.max_retries:
            self.get_taskrunner().notify(_('Max retries exceeded'))
            return

        from gonk.tasks import to_retry

        when = timezone.now() + self.retry_time
        celery_task = to_retry.apply_async(queue=self.queue, kwargs={'task_id': self.id}, eta=when)
        self.celery_id = celery_task.id

        self.save(update_fields=['celery_id'])

    def cancel(self, terminate=False):
        from celery.result import AsyncResult

        AsyncResult(self.celery_id).revoke(terminate=terminate)
        self.status = TaskStatusChoices.CANCELED

        self.save()

    def revert(self):
        from gonk.tasks import to_revert

        celery_task = to_revert.delay(task_id=self.id)
        self.celery_id = celery_task.id

        self.save(update_fields=['celery_id'])

    def expire(self):
        self.get_taskrunner().expire()

    def log_status(self, record, checkpoint: bool = False):
        now = timezone.now().strftime("%Y-%m-%d %H:%M:%S")

        if checkpoint:
            new_line = '{:20} {:70}  Status: {:>15}\n'.format(now, record, self.status)
        else:
            new_line = '{:20} {:70}\n'.format(now, record)

        self.log += new_line

        if checkpoint:
            self.save()
            self.get_taskrunner().notify(data=record)



from django.utils import timezone
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from django.utils.text import gettext_lazy as _

from gonk.models import Task
from gonk.registry import REGISTRY


class TaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = Task
        fields = (
            'id',
            'started_on',
            'finished_on',
            'revert_started_on',
            'revert_finished_on',
            'status',
        )


class RetrieveTaskSerializer(TaskSerializer):
    class Meta:
        model = Task
        fields = TaskSerializer.Meta.fields + (
            'log',
            'results',
        )


class CreateTaskSerializer(serializers.Serializer):
    task_type = serializers.CharField(max_length=255)
    task_input = serializers.JSONField(default={})
    queue = serializers.CharField(max_length=32, default='celery')
    eta = serializers.DateTimeField(default=timezone.now)
    retryable = serializers.BooleanField(default=False)
    max_retries = serializers.IntegerField(default=0)
    retry_seconds = serializers.IntegerField(default=0)

    def validate_task_type(self, value):
        if value not in REGISTRY.registry.keys():
            raise ValidationError(_('No task found for this type'))

        return value

    class Meta:
        fields = ('task_type', 'input', 'queue')



# Generated by Django 3.2.13 on 2022-09-26 15:14

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gonk', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='max_retries',
            field=models.PositiveIntegerField(default=3),
        ),
        migrations.AddField(
            model_name='task',
            name='retries',
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name='task',
            name='retry_time',
            field=models.DurationField(default=datetime.timedelta(seconds=0)),
        ),
        migrations.AddField(
            model_name='task',
            name='retryable',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.CharField(choices=[('ERROR', 'Error'), ('PENDI', 'To DO'), ('DOING', 'Doing'), ('DONE', 'Done'), ('CANCG', 'Cancelling'), ('CANCD', 'Canceled'), ('TOREV', 'To Revert'), ('REVTG', 'Reverting'), ('REVTD', 'Reverted'), ('RTRNG', 'Retrying'), ('RTERR', 'Retry error')], default='PENDI', max_length=5),
        ),
    ]



import logging
import sys
import traceback

from celery import shared_task
from celery.schedules import crontab
from django.utils import timezone

from gonk.beat import add_beat_to_celery
from gonk.models import Task
from gonk.settings import TaskStatusChoices

logger = logging.getLogger(__name__)


def execute(task_id, func):
    try:
        task = Task.objects.get(pk=task_id)
        task.log_status('TASK FOUND')
    except Exception as e:
        logger.error(str(e))
        return

    try:
        func(task)
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        task.status = TaskStatusChoices.ERROR
        task.results = {
            "exception": str(e),
            "line": exc_tb.tb_lineno,
            "traceback": traceback.format_exc()
        }
        task.log_status(f'ERROR: {str(e)}', checkpoint=True)
        logger.error(str(e), exc_info=True)
        task.save()

        # Schedule task for retry if task is retryable
        task.retry()


def runner_func(task):
    task.status = TaskStatusChoices.DOING
    task.started_on = timezone.now()
    task.save()
    task.get_taskrunner().run()
    task.finished_on = timezone.now()
    task.status = TaskStatusChoices.DONE
    task.save()


def reverter_func(task):
    task.status = TaskStatusChoices.REVERTING
    task.revert_started_on = timezone.now()
    task.save()
    task.get_taskrunner().revert()
    task.revert_finished_on = timezone.now()
    task.status = TaskStatusChoices.REVERTED
    task.save()


def retry_func(task):
    task.status = TaskStatusChoices.RETRYING
    task.retries += 1
    task.started_on = timezone.now()
    task.save()
    task.get_taskrunner().retry()
    task.finished_on = timezone.now()
    task.status = TaskStatusChoices.DONE
    task.save()


@shared_task()
def to_run(task_id):
    execute(task_id, runner_func)


@shared_task()
def to_retry(task_id):
    execute(task_id, retry_func)


@shared_task()
def to_revert(task_id):
    execute(task_id, reverter_func)


@shared_task()
def run_schedule(task_type, params: dict = None):
    if params is None:
        params = {}

    Task.create_task(task_type, params)


@shared_task()
def cleanup_gonk_tasks():
    Task.cleanup()


add_beat_to_celery('cleanup_gonk_tasks',
                   'gonk.tasks.cleanup_gonk_tasks',
                   crontab(minute=0, hour=4), [])



from celery import _state as state
from celery.schedules import crontab


def add_beat_to_celery(name: str, task: str, cron: crontab, args):
    state.current_app.conf.beat_schedule.update({
        name: {
            'task': task,
            'schedule': cron,
            'args': args
        }
    })



from django.core.management.base import BaseCommand
from gonk.registry import REGISTRY


class Command(BaseCommand):
    def handle(self, *args, **options):
        available_taskrunners = REGISTRY.registry.keys()

        self.stdout.write(self.style.SUCCESS('Available taskrunners: '))

        for taskrunner in available_taskrunners:
            self.stdout.write(f'- {taskrunner}')



# Generated by Django 3.2.13 on 2022-10-03 10:32

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gonk', '0002_auto_20220926_1514'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='retry_time',
            field=models.DurationField(blank=True, default=datetime.timedelta(0), null=True),
        ),
    ]
